@@ -1,6 +1,6 @@
 /*
- renattach 1.2.0 - Filter that renames/deletes dangerous email attachments
- Copyright (C) 2003  Jem E. Berkes <jberkes at pc-tools dot net>
+ renattach 1.2.1 - Filter that renames/deletes dangerous email attachments
+ Copyright (C) 2003, 2004  Jem E. Berkes <jberkes@pc-tools.net>
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
@@ -43,11 +43,14 @@ int filter_pass(FILE*);
 int header_pass(FILE*);
 int is_line_mime(const char*, int, struct attach*);
 int filter_decision(struct attach*);
-int match_filename(const char*);
+int match_filename(const char*, int);
 int decode_2047(char*, struct namespec*);
 int reencode_name(struct namespec*, const char*, char*);
+void kill_exit(const char*);
 void taking_action(const char*);
+void not_taking_action(const char*, const char*);
 int explain_subject(FILE*, char*);
+void sanitize_shell_cmd(char*);
 void abnormalproc(void);
 
 
@@ -180,7 +183,10 @@ int main(int argc, char** argv)
 						configuration.pipe_cmd[strlen(configuration.pipe_cmd)-1] = '\0';
 				}
 				else
-					configuration.pipe_cmd = optarg;
+				{
+					configuration.pipe_cmd = malloc(1+strlen(optarg));
+					strcpy(configuration.pipe_cmd, optarg);
+				}
 				break;
 
 			case 's':
@@ -211,6 +217,8 @@ int main(int argc, char** argv)
 		configuration.mode = MODE_BAD;		/* default filter mode is badlist */
 	if (!configuration.def_act)
 		configuration.def_act = ACTION_RENAME;	/* default filter action is to rename */
+	if (configuration.pipe_cmd)
+		sanitize_shell_cmd(configuration.pipe_cmd);
 
 	/* Start filter with current action = default action */
 	configuration.action = configuration.def_act;
@@ -269,13 +277,14 @@ int main(int argc, char** argv)
 int filter_pass(FILE* dest)
 {
 	char linebuf[MAXLINEBUF];
-	int onheaders = 1;			/* await end of message headers for delete_exe checks */
+	int onheaders = 1;			/* await end of message headers for kill/delete_exe checks */
 	int on_mime = 0;			/* are we processing a MIME header block? */
 	struct attach m_attachment;		/* for MIME attachments found  */
 	struct attach u_attachment;		/* for uuencoded attachments found */
+	struct zip_info zipstate;		/* for decoding of ZIP files encountered (containing filename) */
 	char uuscanspec[MAXFIELD];		/* safe scanf spec for finding uuencoded attachments */
 	long last_mime_pos = 0;			/* stream position of start of MIME headers */
-	int sigcheckcount = 0;			/* executable signatures are checked for when this is > 0 */
+	int sigcheckcount = 0;			/* for delete_exe and kill_exe, check for sig when > 0 */
 	int skip_encoded = 0;			/*	0=no effect on message
 							1=waiting for blank line
 							2=skipping lines while waiting for blank line/boundary
@@ -284,14 +293,15 @@ int filter_pass(FILE* dest)
 	sprintf(uuscanspec, UUENC_START "%%*d %%%d[^\r\n]", MAXFIELD-1);
 	memset(&m_attachment, 0, sizeof(struct attach));
 	memset(&u_attachment, 0, sizeof(struct attach));
+	init_zip(&zipstate);
 	memset(messageid, 0, sizeof(messageid));
 
 	/* Read line at a time from input */
 	while (fgets(linebuf, sizeof(linebuf), stdin))
 	{
-		/* Watch for end of message headers, if using delete_exe feature */
-		if ((configuration.delete_exe == OPTION_ENABLED) && onheaders &&
-			((*linebuf == '\n') || (*linebuf == '\r')) )
+		/* Watch for end of message headers, if using delete_exe or kill_exe features */
+		if ( ((configuration.delete_exe == OPTION_ENABLED) || (configuration.kill_exe == OPTION_ENABLED))
+			&& onheaders && ((*linebuf == '\n') || (*linebuf == '\r')) )
 		{
 			sigcheckcount = 10;	/* check binary signatures on next few lines */
 			onheaders = 0;
@@ -311,7 +321,7 @@ int filter_pass(FILE* dest)
 				Found a uuencoded attachment, and parsed its filename to the structure.
 				Note that unlike MIME, there are no details, and filtering can be done NOW
 			*/
-			if (configuration.delete_exe == OPTION_ENABLED)
+			if ((configuration.delete_exe==OPTION_ENABLED) || (configuration.kill_exe==OPTION_ENABLED))
 				sigcheckcount = 5;	/* check binary signatures on next few lines */
 			if (filter_decision(&u_attachment))
 			{
@@ -342,7 +352,7 @@ int filter_pass(FILE* dest)
 		{
 			/* Now past end of MIME headers. Decision time, "do we do anything?" */
 			on_mime = 0;
-			if (configuration.delete_exe == OPTION_ENABLED)
+			if ((configuration.delete_exe==OPTION_ENABLED) || (configuration.kill_exe==OPTION_ENABLED))
 				sigcheckcount = 10;	/* check binary signatures on next few lines */
 			if (m_attachment.fattach)
 			{
@@ -376,10 +386,14 @@ int filter_pass(FILE* dest)
 				fprintf(dest, MIME_CDISP " attachment;\n%s\n", nameout);
 				if (*m_attachment.content_enc)
 					fprintf(dest, MIME_CTENC " %s\n", m_attachment.content_enc);
+				/* If user wants to pass Content-ID, add back that header unfiltered */
+				if ((configuration.pass_contentid == OPTION_ENABLED) && *m_attachment.content_id)
+					fprintf(dest, MIME_CNTID " %s\n", m_attachment.content_id);
 			}
 			memset(&m_attachment, 0, sizeof(struct attach));
 		}
 
+		/* Check for encoded executable signatures (delete_exe, kill_exe) */
 		if (sigcheckcount > 0)
 		{
 			sigcheckcount--;
@@ -388,6 +402,11 @@ int filter_pass(FILE* dest)
 				(strncmp(linebuf, B64_EXESIG3, 3) == 0) ||
 				(strncmp(linebuf, B64_EXESIG4, 3) == 0) )
 			{
+				if (configuration.kill_exe == OPTION_ENABLED)
+				{
+					configuration.action = ACTION_KILL;
+					kill_exit(TXT_ENCODEDBODY);	/* kill email and exit */
+				}
 				skip_encoded = 2;
 				configuration.action = ACTION_DELETE;
 				exe_count++;
@@ -398,6 +417,11 @@ int filter_pass(FILE* dest)
 				(strncmp(linebuf+1, UUE_EXESIG3, 3) == 0) ||
 				(strncmp(linebuf+1, UUE_EXESIG4, 3) == 0) )
 			{
+				if (configuration.kill_exe == OPTION_ENABLED)
+				{
+					configuration.action = ACTION_KILL;
+					kill_exit(TXT_ENCODEDBODY);	/* kill email and exit */
+				}
 				skip_encoded = 3;
 				configuration.action = ACTION_DELETE;
 				exe_count++;
@@ -405,6 +429,46 @@ int filter_pass(FILE* dest)
 			}
 		}
 
+		/*
+			Check for filenames inside ZIPs. Any filenames found inside are subject
+			to the standard goodlist/badlist tests and deleted or killed, except that
+				NOTHING IS DONE IF CURRENT ACTION IS 'RENAME' (dummy action)
+		*/
+		if (configuration.search_zip == OPTION_ENABLED)
+		{
+			int decoded;
+			char textbuf[MAXLINEBUF], binbuf[MAXLINEBUF];
+
+			strncpy(textbuf, linebuf, sizeof(textbuf));
+			trim_trailch(textbuf, '\r', '\n', ' ');
+			if (*textbuf == 0)
+				init_zip(&zipstate);
+			else if( (skip_encoded<2) && (decoded=base64_decode_line(textbuf, binbuf)) )
+			{
+				char* zip_pos = binbuf;
+				int available = decoded;
+				while ((zip_pos=unzip_filename(zip_pos, available, &zipstate)))
+				{
+					/* Found a filename inside ZIP archive. Synthesize struct attach */
+					struct attach inzip;
+					memset(&inzip, 0, sizeof(struct attach));
+					inzip.within = 1;
+					strncpy(inzip.ct_fname.name, zipstate.filename, sizeof(inzip.ct_fname.name));
+					if (filter_decision(&inzip))
+					{
+						if (configuration.action == ACTION_DELETE)
+						{
+							skip_encoded = 2;	/* skip rest of attachment */
+							taking_action(zipstate.filename);
+						}
+						else
+							not_taking_action(TXT_INFO_CANTRENZIP, zipstate.filename);
+					}
+					available = (binbuf+decoded) - zip_pos;
+				}
+			}
+		}
+
 		if ( (skip_encoded==1) && ((*linebuf == '\n')||(*linebuf == '\r')) )
 			skip_encoded = 2;	/* now we're actually skipping body lines */
 		else if (skip_encoded == 2)	/* Encoded body lines will be skipped if 'deleting' an attachment. */
@@ -479,10 +543,13 @@ int header_pass(FILE* source)
 				continue;
 		}
 
-		/* Remove existing X- headers specific to RenAttach */
+		/* Keep existing X- headers specific to RenAttach, by renaming as Old- */
 		if (onheaders && ( (strncasecmp(linebuf, HEAD_FILTERED_WITH, sizeof(HEAD_FILTERED_WITH)-1)==0)
 			|| (strncasecmp(linebuf, HEAD_FILTERED_INFO, sizeof(HEAD_FILTERED_INFO)-1)==0) ))
+		{
+			fprintf(dest, "Old-%s", linebuf);
 			continue;
+		}
 
 		/* Append our new headers */
 		if (onheaders && ((*linebuf == '\n') || (*linebuf == '\r')) )
@@ -531,6 +598,7 @@ int header_pass(FILE* source)
 */
 int is_line_mime(const char* line, int on_mime, struct attach* m_attachment)
 {
+	const char* orgline = line;
 	int found_mime = 0;
 
 	/* Is there something worth trying to parse? */
@@ -545,6 +613,7 @@ int is_line_mime(const char* line, int on_mime, struct attach* m_attachment)
 	
 	if (found_mime)
 	{
+		int parsed_name = 0;				/* flag, was name parsed? */
 		char* nameloc = stristr(line, MIME_NAME);	/* case-insensitive search */
 		
 		if (strncasecmp(line, MIME_CTYPE, sizeof(MIME_CTYPE)-1) == 0)
@@ -574,6 +643,17 @@ int is_line_mime(const char* line, int on_mime, struct attach* m_attachment)
 				trim_trailing(m_attachment->content_enc);
 			}
 		}
+                else if (strncasecmp(line, MIME_CNTID, sizeof(MIME_CNTID)-1) == 0)
+		{
+			/* Trying to parse Content-ID */
+			char scanspec[MAXFIELD];
+			sprintf(scanspec, "%%%d[^;\r\n\t]", MAXFIELD-1);	/* generate safe scanf spec */
+			if (sscanf(line+sizeof(MIME_CNTID)-1, scanspec, m_attachment->content_id) == 1)
+			{
+				trim_leading(m_attachment->content_id);
+				trim_trailing(m_attachment->content_id);
+			}
+		}
 		
 		/* Now hunt for filenames! */
 		while (m_attachment->curspec && nameloc)
@@ -604,6 +684,7 @@ int is_line_mime(const char* line, int on_mime, struct attach* m_attachment)
 				(sscanf(nameloc+sizeof(MIME_NAME)-1, scanspec4, parsed) == 1) )
 			{
 				int remaining;
+				parsed_name = 1;
 				line += strlen(parsed);	/* further advance pointer for next search */
 				if (m_attachment->curspec->specformat == FMT_RFC2231)
 					decode_hex(parsed, '%', 0, NULL);
@@ -621,6 +702,18 @@ int is_line_mime(const char* line, int on_mime, struct attach* m_attachment)
 			/* Prepare for next iteration */
 			nameloc = stristr(line, MIME_NAME);	/* case-insensitive search */
 		}
+		
+		/* Might have encountered (obsolete format) line continuation on encoded filename */
+		if (!parsed_name && m_attachment->curspec && (m_attachment->curspec->specformat == FMT_RFC2047))
+		{
+			char parsed[MAXFIELD], scanspec[MAXFIELD];
+			sprintf(scanspec, "%%*[\t ]%%%d[^\r\n\"]", MAXFIELD-1);
+			if ((sscanf(orgline, scanspec, parsed)==1) && (decode_2047(parsed, m_attachment->curspec)==0))
+			{
+				int remaining = MAXFIELD - strlen(m_attachment->curspec->name) - 1;
+				strncat(m_attachment->curspec->name, parsed, remaining);
+			}
+		}
 	}
 	return found_mime;
 }
@@ -648,9 +741,9 @@ int filter_decision(struct attach* m_attach)
 	trim_trailch(m_attach->cd_fname.name, '\t', ' ', '.');
 
 	/* See if either of the parsed filenames gets caught by filter */
-	if (match_filename(m_attach->ct_fname.name))
+	if (match_filename(m_attach->ct_fname.name, m_attach->within))
 		m_attach->curspec = &m_attach->ct_fname;
-	else if (match_filename(m_attach->cd_fname.name))
+	else if (match_filename(m_attach->cd_fname.name, m_attach->within))
 		m_attach->curspec = &m_attach->cd_fname;
 	else
 	{
@@ -666,21 +759,25 @@ int filter_decision(struct attach* m_attach)
 	strcpy(m_attach->curspec->oldname, m_attach->curspec->name);
 
 	if (configuration.action == ACTION_KILL)
-	{
-		/* To properly absorb email, read stdin to completion */
-		char linebuf[MAXLINEBUF];
-		taking_action(m_attach->curspec->oldname);
-		while (fgets(linebuf, sizeof(linebuf), stdin))
-			;
-		exit(excode);	/* if --excode used, indicates filtering happened */
-	}
+		kill_exit(m_attach->curspec->oldname);	/* absorbs email, and exits */
 
-	for (dotsearch = m_attach->curspec->name; *dotsearch; dotsearch++)
+	if (configuration.full_rename == OPTION_ENABLED)
 	{
-		/* replace all periods with underscores */
-		if (*dotsearch == '.')
+		/* Full rename replaces all periods with underscores (safest) */
+		for (dotsearch = m_attach->curspec->name; *dotsearch; dotsearch++)
+		{
+			if (*dotsearch == '.')
+				*dotsearch = '_';
+		}
+	}
+	else
+	{
+		/* Partial rename just replaces final period with underscore */
+		dotsearch = strrchr(m_attach->curspec->name, '.');
+		if (dotsearch)
 			*dotsearch = '_';
 	}
+
 	/*
 		underscores alone make extension safe
 		Also tack on new extension if there is enough buffer space
@@ -699,10 +796,11 @@ int filter_decision(struct attach* m_attach)
 
 /*
 	Does this filename match (by whatever the current criteria are)
+	Provide flag if filename was found inside zip
 	returns 0 - do not act (file poses no risk)
 	returns 1 - do act, something should be done to this file
 */
-int match_filename(const char* filename)
+int match_filename(const char* filename, int inzip)
 {
 	char extension[MAXFIELD];
 	char *lastdot, *listcopy, *token;
@@ -782,10 +880,19 @@ int match_filename(const char* filename)
 		while (token)
 		{
 			int action=0;		/* use default action, unless overridden */
-			char* actswitch = strchr(token, '/');		/* see if action is specified */
+			char *switchar, *actswitch, *inswitch = NULL;
+
+			actswitch = strchr(token, '/');		/* see if action is specified */
 			if (actswitch)
+				inswitch = strchr(actswitch+1, '/');
+			if (inzip && inswitch)			/* if specific action found for inside ZIP */
+				switchar = inswitch;		/* 	use that action */
+			else
+				switchar = actswitch;		/* otherwise, use first action found */
+
+			if (switchar)
 			{
-				switch (actswitch[1])
+				switch (switchar[1])	/* select specific action */
 				{
 					case 'r':
 					case 'R':
@@ -800,8 +907,9 @@ int match_filename(const char* filename)
 						action = ACTION_KILL;
 						break;
 				}
-				*actswitch = '\0';	/* drop the action specification */
 			}
+			if (actswitch)
+				*actswitch = '\0';	/* terminate string at first switch */
 
 			if (strcasecmp(token, extension) == 0)
 			{
@@ -841,7 +949,7 @@ int match_filename(const char* filename)
 /*
 	Attempts to decode an RFC-2047 encoded field
 	Fills in information about fieldspec
-	
+
 	Returns 0 if successful, nonzero otherwise
 */
 int decode_2047(char* field, struct namespec* fieldspec)
@@ -1093,6 +1201,20 @@ int reencode_name(struct namespec* fieldspec, const char* tagbase, char* result)
 }
 
 
+/*
+	Kill this email and exit the program, taking_action along the way
+*/
+void kill_exit(const char* text)
+{
+	/* To properly absorb email, read stdin to completion */
+	char linebuf[MAXLINEBUF];
+	taking_action(text);
+	while (fgets(linebuf, sizeof(linebuf), stdin))
+		;
+	exit(excode);	/* if --excode used, indicates filtering happened */
+}
+
+
 /*
 	Called whenever taking some kind of filtering action (the type: configuration.action)
 	Responsible for counting filtering, logging
@@ -1120,6 +1242,25 @@ void taking_action(const char* caughtfile)
 }
 
 
+/*
+	Called when NOT acting upon filename; logs warning
+	As with taking_action, this resets configuration.action
+*/
+void not_taking_action(const char* reason, const char* missedfile)
+{
+	if (verbose)
+		fprintf(stderr, TXT_INFO_NOTACTING " - %s: %s\n", reason, missedfile);
+	if (configuration.use_syslog == OPTION_ENABLED)
+	{
+		openlog("renattach", LOG_PID, LOG_MAIL);
+		syslog(LOG_WARNING, TXT_INFO_NOTACTING " - %s: %s", reason, missedfile);
+		closelog();
+	}
+	/* Reset to default action */
+	configuration.action = configuration.def_act;
+}
+
+
 /*
 	Explain filtering by writing a new Subject field to dest
 	oldsubject may be NULL (if adding an entirely new Subject field)
@@ -1132,7 +1273,7 @@ int explain_subject(FILE* dest, char* oldsubject)
 	char* therest = "\n";
 	if (oldsubject)
 		therest = oldsubject;
-		
+
 	if (ban_count && *configuration.subj_banned)
 	{
 		if (strcmp(configuration.subj_banned, "#") == 0)
@@ -1175,6 +1316,36 @@ int explain_subject(FILE* dest, char* oldsubject)
 }
 
 
+/*
+	Shell commands are sanitized by stripping out dangerous characters
+*/
+void sanitize_shell_cmd(char* buf)
+{
+	int inpos, outpos=0, length=strlen(buf);
+	char* tmpbuf = calloc(1, 1+length);
+	for (inpos=0; inpos<length; inpos++)
+	{
+		switch (buf[inpos])
+		{
+			case SEMICOLON:
+			case BACKTICK:
+			case AMPERSAND:
+			case DOLLARSIGN:
+			case BACKSLASH:
+			case GREATERTHAN:
+			case LESSTHAN:
+			case VERTICALBAR:
+				break;
+
+			default:
+				tmpbuf[outpos++] = buf[inpos];
+		}
+	}
+	strcpy(buf, tmpbuf);
+	free(tmpbuf);
+}
+
+
 /*
 	SIGPIPE signal handler; clean up tempfile, exit
 */
